// cbun: easy-going implementation of deterministric bunsetsu chunker
//  $Id: cbun.h 605 2011-10-14 12:52:07Z ynaga $
// Copyright (c) 2010-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include <unistd.h>
#include <cstdio>
#include <vector>
#include <string>
#include <mecab.h>
#if defined (USE_DARTS) || defined (USE_DARTS_CLONE)
#include <darts.h>
#elif defined (USE_CEDAR)
#include "cedar.h"
#endif
#include "typedef.h"
#include "opal_conf.h" // opal
#define USE_PRUNE
#define USE_PMT
#include "pa.h"        // opal
#include "pecco.h"    // pecco
#include "jdepp_conf.h"

//  - IN/OUT format: MeCab/KNP; tweak field_t accordingly
//    - labeled data: mark '* ' with the beginning morph of each bunsetsu
//  - classifier:
//    - training: opal (given options are delegated to opal)
//    - testing:  opal (Kd=0) / pecco (provide options after -- in command-line)
//  - performance (on Intel 3.2Ghz CPU; opal options: Kd=2, algo=PA1, ave=0, s=0, c=0.005):
//    + training time: ~14s
//    + chunking accuracy: ~99.7%, compalable with CRF (w/ same feature) or KNP
//    + chunking speed:    ~21000 sent./sec. (almost same as MeCab)

namespace cbun {
  // default parameters
  static const char * CBUN_PECCO_ARGV[] = {"--", "-t", "1", "-i", "4"};
  static const char * MECAB_OPT = "$0 -d /usr/local/Cellar/mecab/0.98/lib/mecab/dic/jumandic"; // correct here
#if   defined USE_JUMAN_POS
  enum field_t {SURF, POS1, POS2, TYPE, INFL, FIN, YOMI, OTHER, NUM_FIELD};
#elif defined USE_IPA_POS
  enum field_t {SURF, POS1, POS2, POS3, POS4, TYPE, INFL, FIN, YOMI, PRON, NUM_FIELD};
#endif
  //
  struct node {
    int          y;
    int          field[NUM_FIELD];
    const char * surface;
    size_t       length;
    const char * feature;
    node (int y_ = -1) : y (y_), surface (0), length (0), feature (0) {};
  };
  // typedef
  class chunker {
  private:
    const opal::option    _opt;
  public:
    std::vector <node> morph; // sentence
    size_t             mlen;   // # morphs
  private:
    opal::mem_pool <opal::ex_t> _ex_pool;
    ny::fv_t              _fv;         // feature vector
    opal::ex_t            _x;
    ny::trie              _trie;       // string bag: string -> int
    ny::uint              _num_lexical_features;
    int                   _unseen;
    ny::uint              _num_lexical_item;
    MeCab::Tagger *       _tagger;
    opal::Model *         _opal;
    pecco::pecco *        _pecco;
    ny::uint              _nf;
    std::string           _dict;        // dda
    std::string           _train;       // event file
    char                  _res[IOBUF_SIZE];
    char*                 _ptr;
    size_t                _n;
    FILE*                 _writer;
    ny::uint _pp, _pn, _np, _nn;
    size_t _ns;
    
    template <opal::option::mode_t MODE>
    void _set_node (const char* surface, const size_t length, const char* feature, int &y) {
      if (mlen >= morph.size ()) morph.resize (mlen + 1);
      node &n = morph[mlen];
      n.y = y; n.surface = surface; n.length = length; n.feature = feature;
      y = -1;
      n.field[SURF] = _trie.exactMatchSearch <ny::trie::result_type> (surface, length);
      const char* cs (feature), * p (cs);
      for (ny::uint i = 1; *cs != '\0' && i < NUM_FIELD; ++i) {
        while (*p != '\0' && *p != ',') ++p; //
        if (i == INFL || i <= POS2) {
          int id = _trie.exactMatchSearch <ny::trie::result_type> (cs, p - cs);
          n.field[i] = id >= 0 ? id : _unseen;
        }
        cs = ++p;
      }
      ++mlen;
    }
    void _set_feature (ny::uint i, bool sort) {
      ny::uint fi = 1;
      _fv.resize (0);
      if (i >= 2) {
        _fv.push_back (fi + morph[i-2].field[SURF]); fi += _num_lexical_features;
        _fv.push_back (fi + morph[i-2].field[POS1]); fi += _num_lexical_features;
        _fv.push_back (fi + morph[i-2].field[POS2]); fi += _num_lexical_features;
        _fv.push_back (fi + morph[i-2].field[INFL]); fi += _num_lexical_features;
      } else {
        fi += 4 * _num_lexical_features;
      }
      _fv.push_back (fi + morph[i-1].field[SURF]); fi += _num_lexical_features;
      _fv.push_back (fi + morph[i-1].field[POS1]); fi += _num_lexical_features;
      _fv.push_back (fi + morph[i-1].field[POS2]); fi += _num_lexical_features;
      _fv.push_back (fi + morph[i-1].field[INFL]); fi += _num_lexical_features;
      _fv.push_back (fi + morph[i].field[SURF]);   fi += _num_lexical_features;
      _fv.push_back (fi + morph[i].field[POS1]);   fi += _num_lexical_features;
      _fv.push_back (fi + morph[i].field[POS2]);   fi += _num_lexical_features;
      _fv.push_back (fi + morph[i].field[INFL]);   fi += _num_lexical_features;
    }
    void _print_ex () const {
      std::fwrite (_x.getLabel () > 0 ? "+1" : "-1", 1, 2, _writer);
      for (opal::fv_it it = _x.begin (); it != _x.end (); ++it)
        std::fprintf (_writer, " %d:1", *it);
      std::fwrite ("\n", 1, 1, _writer);
    }
    template <opal::option::mode_t MODE, typename classifier>
    void _process (classifier* p = 0, bool output = false) {
      ny::uint bid = 0;
      if (MODE == opal::option::TEST) {
        ++_pp;
        if (output) {
          _ptr += std::sprintf (_ptr, "* 0 -1D\n");
          std::memcpy (_ptr, morph[0].surface, morph[0].length);
          _ptr += morph[0].length;
          _ptr += std::sprintf (_ptr, "\t%s\n", morph[0].feature);
        }
      }
      morph[0].y = +1;
      for (ny::uint i = 1; i < mlen; ++i) {
        node &n  = morph[i];
        _set_feature (i, MODE == opal::option::TRAIN);
        if (MODE == opal::option::TRAIN) {
          _opal->set_ex (_x, n.y, _fv, true, _opt.kernel == opal::POLY);
          _ex_pool.put (_x);
          _print_ex ();
        } else {
          const ny::fl_t y_ = p->binClassify (_fv) ? 1 : -1;
          if (y_ >= 0) {
            if (output) _ptr += std::sprintf (_ptr, "* %d -1D\n", ++bid);
            if (n.y > 0) ++_pp; else ++_np;
          } else {
            if (n.y > 0) ++_pn; else ++_nn;
          }
          n.y = y_ >= 0 ? +1 : -1;
          if (output) {
            std::memcpy (_ptr, n.surface, n.length);
            _ptr += n.length;
            _ptr += std::sprintf (_ptr, "\t%s\n", n.feature);
          }
        }
      }
      if (output) _ptr += std::sprintf (_ptr, "EOS\n");
    }
    template <opal::option::mode_t MODE, typename classifier>
    void _analyze_node (classifier * p, const char * surface, const size_t length, const char * feature, int &y, bool output = false) {
      if (*surface == '#') {
        mlen = 0;
        if (MODE == opal::option::TEST) {
          if (output) {
            std::memcpy (_ptr, surface, length); _ptr += length;
            std::memcpy (_ptr, "\n", 1);         _ptr += 1;
          }
        }
      } else if (*surface == '*') {
        y = +1;
      } else if (*surface == 'E' || *feature == 'B') { // EOS
        if (MODE != opal::option::TEST)
          if (++_n % 1000 == 0)
            std::fprintf (stderr, "\rtrain: %ld sent. processed", _n);
        _process <MODE> (p, output);
        if (output) {
          char *check = &_res[0];
          while (check < _ptr) check += write (1, check, _ptr - check);
          _ptr = &_res[0];
        }
      } else {
        _set_node <MODE> (surface, length, feature, y);
      }
    }
    template <opal::option::mode_t MODE, typename classifier>
    void _analyze (const char* fn, classifier * p = 0) {
      std::string mode;
      switch (MODE) {
        case opal::option::TRAIN: mode = "learn";  break;
        case opal::option::TEST:  mode = "chunk"; break;
        default: break;
      }
      FILE* reader = std::strcmp (fn, "_") == 0 ? stdin : std::fopen (fn, "rb");
      if (! reader) ny::print_err (HERE "no such file: %s\n", fn);
      if (reader == stdin)
        std::fprintf (stderr, "(input: STDIN).");
      bool output = MODE == opal::option::TEST && (reader == stdin || _opt.output == 2);
      int  y = -1;
      char buf[IOBUF_SIZE];
      char * q (&buf[0]), * q_end (&buf[0] + (IOBUF_SIZE));
      char * surface (0), * feature (0);
      size_t length  (0), avail (0);
      std::vector <char *> pos;
      while ((avail = std::fread (q, sizeof (char), q_end - q, reader))) {
        q_end = q + avail;
        q = &buf[0];
        while (1) {
          pos.resize (0);
          for (char * r = q; r + 3 < q_end; ++r) {
            pos.push_back (r);
            if (*r == 'E') { q = r + 4; break; }  // find EOS\n
            while (r != q_end && *r != '\n') ++r; // next line
          }
          if (pos.empty () || *pos.back () != 'E') { // premature input
            std::memmove (&buf[0], q, q_end - q);
            q     = &buf[0] + (q_end - q);
            q_end = &buf[0] + IOBUF_SIZE;
            break;
          }
          for (size_t i = 0; i < pos.size (); ++i) {
            char * r = surface = pos[i];
            size_t read = (i == pos.size () - 1? q : pos[i + 1]) - pos[i];
            r[read - 1] = '\0';
            while (*r != '\0' && *r != '\t') ++r;
            length  = r - surface;
            feature = *r == '\t' ? ++r : r;
            _analyze_node <MODE> (p, surface, length, feature, y, output);
          }
        }
      }
      if (MODE != opal::option::TEST)
        std::fprintf (stderr, "\rtrain: %ld sent. processed; done.\n", _n);
      if (reader != stdin) std::fclose (reader);
    }
  public:
    chunker (const opal::option &opt) : _opt (opt), morph (), mlen (0), _ex_pool (), _fv (), _x (), _trie (), _num_lexical_features (0), _unseen (0), _tagger (0), _opal (0), _pecco (0), _nf (0), _dict (), _train (std::string (_opt.model) + ".train"), _res (), _ptr (&_res[0]), _n (0), _writer (0), _pp (0), _pn (0), _np (0), _nn (0), _ns (0)
    {
      size_t pos = _train.rfind ("/");
      _dict = _train.substr (0, pos == std::string::npos ? 0 : pos + 1) + "dic";
    };
    ~chunker () {
      if (_opal)   delete _opal;
      if (_pecco)  delete _pecco;
      if (_tagger) delete _tagger;
    }
    void train () {
      FILE * fp = std::fopen (_dict.c_str (), "rb");
      std::fread (&_num_lexical_features, sizeof (ny::uint), 1, fp);
      _unseen = ++_num_lexical_features - 1;
      std::fclose (fp);
      _trie.open (_dict.c_str (), "rb", sizeof (ny::uint) * 2);
      _opal = new opal::Model (_opt);
      _writer = std::fopen (_train.c_str (), "w");
      if (! _writer) {
        std::fprintf (stderr, "cannot write to file: %s\n", _train.c_str ());
        std::exit (1);
      }
      _analyze <opal::option::TRAIN> (_opt.train, _opal);
      std::fclose (_writer);
      _opal->train (_ex_pool, _opt.iter);
      _opal->save (_opt.model);
      delete _opal;
      _opal = 0;
    }
    void test (pecco::option &pecco_opt) {
      setup (pecco_opt);
      _analyze <opal::option::TEST> (_opt.test, _pecco);
      std::fprintf (stderr, "acc. %2.3f%% (pp %d) (pn %d) (np %d) (nn %d)\n",
                    (_pp+_nn) * 100.0 / (_pp+_pn+_np+_nn), _pp, _pn, _np, _nn);
    }
    void setup (pecco::option &pecco_opt) {
      _tagger = MeCab::createTagger (MECAB_OPT);
      FILE * fp = std::fopen (_dict.c_str (), "rb");
      std::fread (&_num_lexical_features, sizeof (ny::uint), 1, fp);
      _unseen = ++_num_lexical_features - 1;
      std::fclose (fp);
      _trie.open (_dict.c_str (), "rb", sizeof (ny::uint) * 2);
      if (_opt.kernel == opal::LINEAR) {
        _pecco = new pecco::pecco (_opt, reinterpret_cast <opal::Model*> (0));
        _pecco->load (_opt.model);
      } else { // delegate polynomial kernel computation to pecco
        pecco_opt.model = _opt.model;
        pecco_opt.train = _train.c_str ();
        pecco_opt.event = _train.c_str ();
        _pecco = new pecco::pecco (pecco_opt,
                                   reinterpret_cast <pecco::kernel_model*> (0));
        _pecco->load (pecco_opt.model);
      }
    }
    void chunk (char * line, size_t len) {
      int y = 0;
      mlen = 0;
      if (_opt.kernel == opal::LINEAR) {
        for (MeCab::Node * n = _tagger->parseToNode (line, len)->next;
             n; n = n->next)
          _analyze_node <opal::option::TEST> (_opal, n->surface, n->length, n->feature, y);
      } else {
        for (MeCab::Node * n = _tagger->parseToNode (line, len)->next;
             n; n = n->next)
          _analyze_node <opal::option::TEST> (_pecco, n->surface, n->length, n->feature, y);
      }
    }
  };
}
