// J.DepP -- Japanese Dependency Parsers
//  $Id: pdep.cc 762 2012-01-04 15:37:59Z ynaga $
// Copyright (c) 2008-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#include "pdep.h"

namespace pdep {

  static const char * input0[] = { "raw", "chunk", "depnd" };
  
  void parser::_add_context_feature (const int from, const int to) {
    // assuming _context_feature_bits to be cleared
    for (int k = from; k <= to; ++k) // merge feature bits;
      for (size_t i = 0; i < _context_feature_bits.size (); ++i)
        _context_feature_bits[i] |= _s->bun[k].fbits[i];
    // set context features by twiddling bits
    //   cf. http://www-cs-faculty.stanford.edu/~uno/fasc1a.ps.gz (p. 8)
    for (flag_t::iterator it = _context_feature_bits.begin ();
         it != _context_feature_bits.end (); _fi += FLAG_LEN, ++it)
      while (*it) {
        pecco::byte_4 b (static_cast <float> (*it & - *it)); // pull rightmost 1
        _fv.push_back (_fi + (b.u >> 23) - 127); // pick it via floating number
        *it &= (*it - 1); // unset rightmost 1
      }
  }
  inline void parser::_add_boolean_feature (const bool flag)
  { _fv.push_back (flag ? _fi : _fi + 1); _fi += 2; }
  inline void parser::_add_boolean_feature (const bool flag, const bool flag_)
  { _fv.push_back (flag ? _fi : (flag_ ? _fi + 1 : _fi + 2)); _fi += 3; }
  void parser::_add_global_feature (const int i, const int j, const Bunsetsu &bi, const Bunsetsu &bj) {
    // activate global feature
#ifndef USE_EMNLP_FEAT
    int  bracket_in = 0;
#endif
    bool period_in (false), comma_in (false);
    //
    for (int k = i + 1; k < j; ++k) {
      const Bunsetsu &b = _s->bun[k];
#ifndef USE_EMNLP_FEAT
      bracket_in += b.bracket;
#endif
      period_in  |= b.period;
      comma_in   |= b.comma;
    }
    // base; 0.9163; 0.5549
    // mod-head: distance / parenthesis / punctuation
    _add_boolean_feature (j - i == 1, j - i > 6);           // 0.9153; 0.5506
#ifndef USE_EMNLP_FEAT
    _add_boolean_feature (bracket_in == 0, bracket_in > 0); // 0.9093; 0.5355
#endif
    _add_boolean_feature (! period_in); // 0.9164; 0.5543
    _add_boolean_feature (! comma_in);  // 0.9157; 0.5527
    // mod: bos  / parenthesis / punctuation
    _add_boolean_feature (i == 0); // 0.9159; 0.5551
    _add_boolean_feature (bi.bracket == 0, bi.bracket > 0); // 0.9163; 0.5560
    _add_boolean_feature (! bi.period); // 0.9143; 0.5485
    _add_boolean_feature (! bi.comma);  // 0.9091; 0.5328
    // head: eos / parenthesis / punctuation
    _add_boolean_feature (j == _s->blen - 1); // 0.9164; 0.5558
    _add_boolean_feature (bj.bracket == 0, bj.bracket > 0); // 0.9109; 0.5384
    _add_boolean_feature (! bj.period); // 0.9141; 0.5496
    _add_boolean_feature (! bj.comma);  // 0.9144; 0.5510
    // particle
    _add_context_feature (i + 1, j - 1); // 0.9140; 0.5485
#ifndef USE_EMNLP_FEAT
    _add_context_feature (i, i); // 0.9148; 0.5516
    _add_context_feature (j, j); // 0.9160; 0.5527
    // _fi = 681101
#ifdef USE_STACKING
    _add_boolean_feature (bi.cand == -1);
    _add_boolean_feature (bi.cand ==  j);
#endif
#endif
  }
  inline void parser::_add_string_feature (const int &id)
  { _fv.push_back (_fi + id); _fi += _dict->num_lexical_features (); }
  inline void parser::_add_string_feature (const int &id, const bool flag)
  { if (flag) _fv.push_back (_fi + id); _fi += _dict->num_lexical_features (); }
  //
  void parser::_add_cluster_feature (const Morph &m) {
    for (ny::uint i = 0; i < _opt.clen; ++i)
      _add_string_feature (m.cluster (i), (1 << i) & _opt.cbits);
  }
  void parser::_add_cluster_feature (const Bunsetsu &b, const Morph * ms) {
    for (ny::uint i = 0; i < _opt.clen; ++i) {
      if ((1 << i) & _opt.cbits) {
        _add_string_feature (ms[0].cluster (i));
        _add_string_feature (ms[b.head].cluster (i), b.head != -1);
        _add_string_feature (ms[b.tail].cluster (i), b.tail != -1);
        _add_string_feature (ms[b.mlen - 1].cluster (i));
      } else {
        _fi += 4 * _dict->num_lexical_features ();
      }
    }
  }
  // set features in a particlular morpheme
  void parser::_add_morpheme_feature (const int idx, const int offset) {
    const Morph &m = _s->morph[idx + offset];
    _add_string_feature (m.surf ()); // 0.9064; 0.5243
    _add_string_feature (m.pos1 ()); // 0.9148; 0.5500
    _add_string_feature (m.pos2 ()); // 0.9134; 0.5442
    _add_string_feature (m.infl ()); // 0.9111; 0.5349
    if (_opt.clen) _add_cluster_feature (m);
  }
  inline void parser::_add_morpheme_feature (const int idx, const int offset, const bool flag) {
    if (flag) _add_morpheme_feature (offset, idx);
    else      _fi += (4 + _opt.clen) * _dict->num_lexical_features ();
  }
  void parser::_add_lexical_feature (const Bunsetsu &b) {
    const Morph * ms = &_s->morph[b.mpos];
    _add_string_feature (ms[0].surf ());                    // 0.9156; 0.5537
    _add_string_feature (ms[b.head].surf (), b.head != -1); // 0.9163; 0.5529
    _add_string_feature (ms[b.tail].surf (), b.tail != -1); // 0.9158; 0.5529
    _add_string_feature (ms[b.mlen - 1].surf ());           // 0.9162; 0.5545
    if (_opt.clen) _add_cluster_feature (b, ms);
  }
  inline void parser::_add_lexical_feature (const Bunsetsu &b, bool flag) {
    if (flag) _add_lexical_feature (b);
    else      _fi += (4 + _opt.clen) * _dict->num_lexical_features ();
  }
  void parser::_event_gen_from_tuple (const int i) { // for chunking
    _fi = 1;
    _fv.clear ();
    _add_morpheme_feature (0, i - 2, i - 2 >= 0);
    _add_morpheme_feature (0, i - 1);
    _add_morpheme_feature (0, i);
  }
  void parser::_event_gen_from_tuple (const int i, const int j) {
    _fi = 1;
    _fv.clear ();
    const Bunsetsu &bi (_s->bun[i]), &bj (_s->bun[j]);
    // _add_morpheme_feature (bi.mpos, 0); // 0.9161; 0.5572
    _add_morpheme_feature (bi.mpos, bi.head, bi.head >= 0);  // 0.9125; 0.5429
    _add_morpheme_feature (bi.mpos, bi.tail, bi.tail >= 0);  // 0.9108; 0.5340
    // _add_morpheme_feature (bi.mpos, bi.mlen - 1);  // 0.9167; 0.5566
#ifndef USE_EMNLP_FEAT
    _add_morpheme_feature (bj.mpos, 0); // 0.9151; 0.5553
#endif
    _add_morpheme_feature (bj.mpos, bj.head, bj.head >= 0);  // 0.9121; 0.5392
    _add_morpheme_feature (bj.mpos, bj.tail, bj.tail >= 0);  // 0.9143; 0.5454
    // _add_morpheme_feature (bj.mpos, bj.mlen - 1);  // 0.9167; 0.5556
#ifndef USE_EMNLP_FEAT
    _add_lexical_feature (_s->bun[j + 1], j < _s->blen - 1); // 0.9099; 0.5363
    // _add_lexical_feature (_s->bun[i - 1], i > 0);     // 0.9172; 0.5582
    // _add_lexical_feature (_s->bun[j - 1], j - 1 > i); // 0.9168; 0.5568
#endif
    _add_global_feature (i, j, bi, bj); // 0.8806; 0.4651
#if ! defined (USE_EMNLP_FEAT) && defined (USE_STACKING)
    _add_lexical_feature (_s->bun[_s->bun[i].cand], j != _s->bun[i].cand);
#endif
    // add new features (>= _fi)  here
  }
  void parser::_event_gen_from_tuple (const int i, const int j, const int k) {
    _fi = 1;
    _fv.clear ();
    const Bunsetsu &bi (_s->bun[i]), &bj (_s->bun[j]), &bk (_s->bun[k]);
    _add_morpheme_feature (bi.mpos, bi.head, bi.head >= 0);
    _add_morpheme_feature (bi.mpos, bi.tail, bi.tail >= 0);
    _add_morpheme_feature (bj.mpos, 0);
    _add_morpheme_feature (bj.mpos, bj.head, bj.head >= 0);
    _add_morpheme_feature (bj.mpos, bj.tail, bj.tail >= 0);
    _add_morpheme_feature (bk.mpos, 0);
    _add_morpheme_feature (bk.mpos, bk.head, bk.head >= 0);
    _add_morpheme_feature (bk.mpos, bk.tail, bk.tail >= 0);
#ifndef USE_EMNLP_FEAT
    _add_lexical_feature (_s->bun[j + 1], j < _s->blen - 1); // A07
    _add_lexical_feature (_s->bun[k + 1], k < _s->blen - 1); // A07
#endif
    _add_global_feature (i, j, bi, bj);
    _add_global_feature (i, k, bi, bk);
#if ! defined (USE_EMNLP_FEAT) && defined (USE_STACKING)
    _add_lexical_feature (_s->bun[_s->bun[i].cand], j != _s->bun[i].cand);
#endif
    // add new features (>= _fi)  here
  }

#if defined (USE_OPAL) || defined (USE_MAXENT)
  inline void parser::_processSample (const bool flag) {
    if (_opt.learner == OPAL) {
#ifdef USE_OPAL
      opal::ex_t x;
      _opal->set_ex (x, flag ? +1 : -1, _fv, true, _opal_opt.kernel == opal::POLY);
      _ex_pool.put (x);
#endif
    } else if (_opt.learner == MAXENT) {
#ifdef USE_MAXENT
      static ME_Sample ms;
      ms.label = flag ? "+1" : "-1";
      ms.features.clear ();
      _project (ms);
      _libme->add_training_sample (ms);
#endif
    }
  }
#endif
  template <const process_t MODE>
  void parser::_parseLinear () {
    static std::stack <ny::uint> stack;
    const int len = _s->blen;
    for (int j = 1; j < len; ++j) {
      stack.push (j - 1);
      while (! stack.empty ()) {
        const int i = stack.top ();
        if (j != len - 1) {
          _event_gen_from_tuple (i, j);
          bool flag = (j == _s->bun[i].did);
          // output example for training / fstrie construction
          if (MODE != PARSE) _print_ex (flag);
          if (MODE != LEARN) flag = _pecco->binClassify (_fv); // parse
#if defined (USE_OPAL) || defined (USE_MAXENT)
          else               _processSample (flag);     // learn
#endif
          if (! flag) break;
        }
        _s->bun[i].est_did = j;
        stack.pop ();
      }
    }
  }
  template <const process_t MODE>
  void parser::_parseChunking () {
    static std::list <chunk_info> cinfo;
    const int len = _s->blen;
    for (int i = 0; i < len - 1; ++i)
      cinfo.push_back (chunk_info (i));
    // cinfo.push_back (chunk_info (i, false)); // deterministic
    while (! cinfo.empty ()) {
      typename std::list <chunk_info>::reverse_iterator it = cinfo.rbegin ();
      bool next = true; // D
      int i = -1;
      int j = it->id;
      _s->bun[j].est_did = len - 1;
      for (++it; it != cinfo.rend (); ++it) {
        i = it->id;
        // if (! it->done) { // deterministic (assuming no dynamic features)
        _event_gen_from_tuple (i, j);
        bool flag = (j == _s->bun[i].did);
        if (MODE != PARSE) _print_ex (flag);
        if (MODE != LEARN) flag = _pecco->binClassify (_fv);
#if defined (USE_OPAL) || defined (USE_MAXENT)
        else               _processSample (flag);
#endif
        if (flag) _s->bun[i].est_did = j;
        // it->done = true; // deterministic
        // }
        if (_s->bun[i].est_did == -1 && next) { // O & D
          it = std::list <chunk_info>::reverse_iterator (cinfo.erase (it.base ()));
          // it->done = false; // deterministic
        }
        next = _s->bun[i].est_did != -1;
        j = i;
      }
      if (_s->bun[i].est_did != -1) cinfo.erase (it.base ());
    }
  }
  template <const process_t MODE>
  void parser::_parseBackward () {
    const int len = _s->blen;
    for (int i = len - 2; i >= 0; --i) {
      ny::fl_t max_prob = 0;
      for (int j = i + 1; j != len - 1; j = _s->bun[j].est_did) { // deterministic
        ny::fl_t prob = 0;
        _event_gen_from_tuple (i, j);
        bool flag = (j == _s->bun[i].did);
        // output example for training / fstrie construction
        if (MODE != PARSE) _print_ex (flag);
        if (MODE == LEARN) { // learn
#if defined (USE_OPAL) || defined (USE_MAXENT)
          _processSample (flag);
#endif
          if (flag) { _s->bun[i].est_did = j; prob = 1.0; }
          // { _s->bun[i].est_did = j; prob = 1.0; break; } // deterministic
        } else {
          prob = _pecco->getProbability (_fv);
        }
        // if (prob >= 0.5) { _s->bun[i].est_did = j; break; } // deterministic
        if (prob > max_prob) { _s->bun[i].est_did = j; max_prob = prob; }
      }
      // if (_s->bun[i].est_did == -1) _s->bun[i].est_did = len - 1;
    }
  }
  template <const process_t MODE>
  void parser::_parseTournament () {
    const int len = _s->blen;
    if (MODE == LEARN) { // I don't merge the two outermost loops
      for (int i = 0; i < len - 2; ++i) {
        int h = _s->bun[i].did; // head
        for (int j = i + 1; j <= len - 1; ++j) {
          bool flag = true;
          if      (j < h) { flag = true;  _event_gen_from_tuple (i, j, h); }
          else if (j > h) { flag = false; _event_gen_from_tuple (i, h, j); }
          else continue;
#if defined (USE_OPAL) || defined (USE_MAXENT)
          _processSample (flag);
#endif
          _print_ex (flag);
        }
      }
    } else {
      for (int i = len - 2; i >= 0; --i) {
        int did = i + 1; // head
        int j = _s->bun[did].est_did; // head of head
        while (j != -1) {
          _event_gen_from_tuple (i, did, j);
          if (MODE == CACHE) _print_ex (did < _s->bun[i].did);
          bool flag = _pecco->binClassify (_fv);
          if (flag) did = j; // RIGHT
          j = _s->bun[j].est_did;
        }
        _s->bun[i].est_did = did;
      }
    }
  }
  template <const process_t MODE>
  void parser::_parse () {
    TIMER (if (MODE != LEARN) _depnd_t->startTimer ());
    if (MODE != LEARN) _switch_classifier (DEPND);
    switch (_opt.parser) {
      case LINEAR:     _parseLinear     <MODE> (); break;
      case CHUNKING:   _parseChunking   <MODE> (); break;
      case TOURNAMENT: _parseTournament <MODE> (); break;
      case BACKWARD:   _parseBackward   <MODE> (); break;
    }
    TIMER (if (MODE != LEARN) _depnd_t->stopTimer ());
    if (MODE == PARSE && _opt.input != RAW) _collectStat <DEPND> ();
  }
  template <const process_t MODE>
  inline void parser::_chunk () {
    TIMER (if (MODE != LEARN) _chunk_t->startTimer ());
    if (MODE != LEARN) _switch_classifier (CHUNK);
    _s->addBunsetsu (0);
    for (int i (1), bi (1); i < _s->mlen; ++i) {
      _event_gen_from_tuple (i);
      bool flag = false;
      if (MODE != PARSE) {
        if (i == _s->bun[bi].mpos) { ++bi; flag = true; }
        // output example for training / fstrie construction
        _print_ex (flag);
      }
      if (MODE != LEARN) { if (_pecco->binClassify (_fv)) _s->addBunsetsu (i); }
#if defined (USE_OPAL) || defined (USE_MAXENT)
      else               _processSample (flag);
#endif
    }
    TIMER (if (MODE != LEARN) _chunk_t->stopTimer ());
    if (MODE == PARSE && _opt.input != RAW) _collectStat <CHUNK> ();
  }
  template <>
  void parser::_collectStat <CHUNK> () {
    ++_chunk_stat.snum;
    Bunsetsu * const bs = _s->bun;
    int blen = 1; while (bs[blen].mpos) ++blen;
    int pp (0), pn (0), np (0);
    for (int i (1), j (blen + 1);;) {
      if (i == blen && j == _s->blen) { ++pp; break; }
      if (i == blen || j == _s->blen)
        { pn += blen - i + 1; np += _s->blen - j + 1; break; }
      if (bs[i].mpos == bs[j].mpos) {
        if (bs[i - 1].mpos == bs[j - 1].mpos) ++pp; else { ++pn; ++np; }
        ++i; ++j;
      } else
        if (bs[i].mpos < bs[j].mpos) { ++i; ++pn; } else { ++j; ++np; }
    }
    if (pp == blen) ++_chunk_stat.scorr;
    _chunk_stat.pp += pp; _chunk_stat.pn += pn; _chunk_stat.np += np;
    // dirty bit
    for (int i (0), j (blen); j < _s->blen; ++i, ++j)
      _s->bun[i].mpos = _s->bun[j].mpos;
    _s->blen = _s->blen - blen;
    _s->setup (_dict);
  }
  template <> void parser::_collectStat <DEPND> () {
    if (_s->blen >= 1) { // set >= 2 to ignore sentene w/ a single bunsetsu
      ++_depnd_stat.snum;
      _depnd_stat.bnum += _s->blen - 1;
      int bcorr = 0;
      const int len = _s->blen;
      for (int i = 0; i < len - 1; ++i)
        if (_s->bun[i].est_did == _s->bun[i].did) ++bcorr;
      // std::printf ("%d\n", b.est_did == b.did ? 1 : 0); // for McNemar
      _depnd_stat.bcorr += bcorr;
      if (bcorr == len - 1) ++_depnd_stat.scorr;
      // std::printf ("%d\n", _flag & 0x2 ? 1 : 0); // for McNemar
    }
  }
  template <const process_t MODE>
  void parser::_analyze () {
    setSentence ();
    size_t n = 0;
#ifdef USE_AS_STANDALONE
    if (_opt.input == RAW) {
      std::string mecab_opt ("$0");
      if (_opt.mecab_dic) mecab_opt += std::string (" -d ") + _opt.mecab_dic;
      MeCab::Tagger * tagger = MeCab::createTagger (mecab_opt.c_str ());
      if (! tagger)
        ny::print_err (HERE "fail to invoke MeCab; you may want to set [-d].\n");
      char header[1024];
      char * line = 0;
      size_t read = 0;
      while (1) {
        TIMER (_io_t->startTimer ());
        const bool input_ok  = ny::getLine (stdin, line, read);
        TIMER (_io_t->stopTimer ());
        if (! input_ok) break;
        TIMER (_preproc_t->startTimer ());
        size_t header_len = std::sprintf (header, "# S-ID: %ld; J.DepP\n", ++n);
        _s->setHeader (header, header_len);
        for (MeCab::Node * m = tagger->parseToNode (line, read - 1)->next;
             m->stat != MECAB_EOS_NODE; m = m->next) // set_feature
          _s->addMorph (m->surface, m->length, m->feature, _dict);
        TIMER (_preproc_t->stopTimer ());
        _chunk <PARSE> ();
        _s->setup (_dict); // bug?
        _parse <PARSE> ();
        TIMER (_io_t->startTimer ());
        _s->print ();
        TIMER (_io_t->stopTimer  ());
        _s->clear ();
      }
      std::fclose (stdin);
      delete tagger;
    } else {
#endif
      static const char * mode[] = { "learn", "parse", "both", "cache" };
      int fd = 0; // stdin
      if (MODE == PARSE)
        std::fprintf (stderr, "(input: STDIN)\n");
      else {
        if ((fd = open (_opt.train, O_RDONLY)) == -1)
          ny::print_err (HERE "no such file: %s\n", _opt.train);
        std::string model (_opt.model_dir + "/" + input0[_opt.input]);
        if (_opt.input == DEPND) {
          char sigparse[16]; std::sprintf (sigparse, ".p%d", _opt.parser);
          model += sigparse;
        }
        if (MODE == LEARN) {
          const std::string train (model + ".train");
          _writer = std::fopen (train.c_str (), "wb");
        } else {
          std::string event (model + ".event");
          if (_opt.learner == SVM)
            event += std::string (".s") + _pecco_opt.sigma;
          _writer = std::fopen (event.c_str (), "wb");
        }
      }
      const bool output = _opt.input == RAW ||
                          (MODE == PARSE && _opt.verbose == -1);
      char buf[IOBUF_SIZE];
      char * q (&buf[0]), * q_end (&buf[0] + IOBUF_SIZE);
      ssize_t avail = 0;
      std::vector <char *> pos;
      while (1) {
        TIMER (if (MODE != LEARN) _io_t->startTimer ());
        const bool input_ok = (avail = read (fd, q, q_end - q)) > 0;
        TIMER (if (MODE != LEARN) _io_t->stopTimer ());
        if (! input_ok) break;
        q_end = q + avail;
        q = &buf[0];
        while (1) {
          TIMER (if (MODE != LEARN) _io_t->startTimer ());
          // read input
          pos.clear ();
          for (char * r = q; r + 3 < q_end; ++r) {
            pos.push_back (r);
            if (*r == 'E') { q = r + 4; break; }  // find EOS\n
            while (r != q_end && *r != '\n') ++r; // next line
          }
          if (pos.empty () || *pos.back () != 'E') { // premature input
            std::memmove (&buf[0], q, q_end - q);
            q     = &buf[0] + (q_end - q);
            q_end = &buf[0] + IOBUF_SIZE;
            break;
          }
          TIMER (if (MODE != LEARN) _io_t->stopTimer ());
          // process
          TIMER (if (MODE != LEARN) _preproc_t->startTimer ());
          _s->clear ();
          ++n;
          if (_opt.input == RAW) {
            char header[1024];
            size_t header_len = std::sprintf (header, "# S-ID: %ld; J.DepP\n", n);
            _s->setHeader (header, header_len);
          }
          for (size_t i = 0; i < pos.size () - 1; ++i) {
            char * line = pos[i];
            const size_t read = pos[i + 1] - pos[i];
            if (_opt.input == RAW)
              _s->addMorph (line, read - 1, _dict);
            else
              switch (*line) {
                case '*': _s->addBunsetsu (line, read - 1, _s->mlen); break;
                case '#': if (output) _s->setHeader (line, read);     break;
                default:  _s->addMorph (line, read - 1, _dict);
              }
          }
          TIMER (if (MODE != LEARN) _preproc_t->stopTimer ());
          if (_opt.input != DEPND) _chunk <MODE> ();
          if (_opt.input != CHUNK) _s->setup (_dict); // bug?
          if (_opt.input != CHUNK) _parse <MODE> ();
          TIMER (if (MODE != LEARN) _io_t->startTimer ());
          if (output) _s->print ();
          TIMER (if (MODE != LEARN) _io_t->stopTimer ());
          if (MODE != PARSE && n % 1000 == 0)
            std::fprintf (stderr, "\r%s: %ld sent. processed", mode[MODE], n);
          if (MODE != PARSE && _opt.max_sent && n >= _opt.max_sent)
            { lseek (fd, 0, SEEK_END); break; } // go to the end
        }
      }
      close (fd);
      if (avail == 0 && q == q_end)
        ny::print_err ("set a larger value to IOBUF_SIZE.\n");
      if (MODE != PARSE) {
        std::fprintf (stderr, "\r%s: %ld sent. processed.", mode[MODE], n);
        std::fclose (_writer);
      }
      std::fprintf (stderr, "\n");
#ifdef USE_AS_STANDALONE
    }
#endif
    delete _s;
  }
  void parser::_learn () {
    std::string model (_opt.model_dir + "/" + input0[_opt.input]);
    if (_opt.input == DEPND) {
      char sigparse[16]; std::sprintf (sigparse, ".p%d", _opt.parser);
      model += sigparse;
    }
    switch (_opt.learner) { // they should be newed earlier
#ifdef USE_OPAL
      case OPAL:
        _opal->train (_ex_pool, _opal_opt.iter);
        _opal->save (model.c_str ());
        break;
#endif
#ifdef USE_SVM
      case SVM:
        {
          const std::string train (model + ".train");
          TinySVM::Example * ex = new TinySVM::Example;
          ex->read (train.c_str ());
          _tinysvm = ex->learn (_tiny_param);
          _tinysvm->write (model.c_str ());
          delete ex;
          break;
        }
#endif
#ifdef USE_MAXENT
      case MAXENT:
        switch (_maxent_opt.algo) {
          case SGD:   _libme->use_SGD ();
          case OWLQN: _libme->use_l1_regularizer (_maxent_opt.reg_cost); break;
          case LBFGS: _libme->use_l2_regularizer (_maxent_opt.reg_cost); break;
          default: assert (! "MAXENT optimizer disabled");
        }
        _libme->train ();
        _libme->save_to_file (model);
        break;
#endif
      default: assert (! "learner disabled");
    }
  }
  void parser::_registerMorph (char * cs, const ssize_t &len, sbag_t &sbag,
                               std::set <ny::uint> &context_feature_ids) {
    ny::uint surf (0), i (0);
    for (char * p (cs), * const p_end (p + len);
         cs < p_end && i < NUM_FIELD; cs = ++p, ++i) {
      if (i == SURF) while (p != p_end && *p != SURFACE_END) ++p;
      else           while (p != p_end && *p != FEATURE_SEP)  ++p;
      *p = '\0';
      // if (i <= POS2 || i == INFL) {
      if (i == SURF || i == POS1 || i == POS2 || i == INFL) {
        sbag_t::iterator it = sbag.find (cs);
        if (it == sbag.end ()) { // new entry
          char * copy = new char[p - cs + 1];
          std::strcpy (copy, cs);
          it = sbag.insert (sbag_t::value_type (copy, sbag.size ())).first;
        }
        const ny::uint id = it->second;
        switch (i) {
          case SURF: surf = id; break;
          case POS1:
            if (std::strcmp (cs, POST_PARTICLE) == 0)
              context_feature_ids.insert (surf);
            break;
#ifdef USE_EMNLP_FEAT
          case POS2:
            if (std::strcmp (cs, BRACKET_START) == 0 ||
                std::strcmp (cs, BRACKET_END)   == 0)
              context_feature_ids.insert (surf);
            break;
#endif
          default: break;
        }
      }
    }
#ifdef USE_JUMAN_POS
    if (i < NUM_FIELD)
      ny::print_err (HERE "# fields is less than %d.\n", NUM_FIELD);
#endif
  }
  // morphological dictionary are extracted from the training data
  void parser::_setMorphDic () {
    if (_opt.verbose > 0) std::fprintf (stderr, "Loading dict..");
    std::string dict (_opt.model_dir + "/dic");
    struct stat st;
    if (stat (dict.c_str (), &st) != 0) {
      if (_opt.verbose > 0)
        std::fprintf (stderr, "not found; reading %s..", _opt.train);
      sbag_t              sbag;
      std::set <ny::uint> context_feature_ids;
      FILE * reader = std::fopen (_opt.train, "r");
      if (! reader) ny::print_err (HERE "no such file: %s\n", _opt.train);
      char * line = 0;
      size_t read = 0;
      while (ny::getLine (reader, line, read))
        if (*line != '*' && *line != '#' && *line != 'E')
          _registerMorph (line, read - 1, sbag, context_feature_ids);
      std::fclose (reader);
      if (_opt.verbose > 0)
        std::fprintf (stderr, "done.\n");
      const ny::uint num_lexical_features = sbag.size ();
      const ny::uint num_context_features = context_feature_ids.size ();
      if (num_context_features == 0)
        ny::print_err (HERE "no particles found in %s\n"
                       "\tthe charset / posset may mismatch with jdepp_conf.h\n",
                       _opt.train);
      FILE * writer = std::fopen (dict.c_str (), "wb");
      std::fwrite (&num_lexical_features, sizeof (ny::uint), 1, writer);
      std::fwrite (&num_context_features, sizeof (ny::uint), 1, writer);
      std::fclose (writer);
      //
      std::vector <const char *> str;
      std::vector <size_t>       len;
      std::vector <int>          val;
      for (sbag_t::const_iterator it = sbag.begin (); it != sbag.end (); ++it) {
        str.push_back (it->first);
        len.push_back (std::strlen (it->first));
        // move (gap) lexical featuress forward and assign smaller IDs
        const ny::uint id = it->second;
        std::set <ny::uint>::iterator jt = context_feature_ids.lower_bound (id);
        const ny::uint index = std::distance (context_feature_ids.begin (), jt);
        val.push_back (*jt == id ? index: id + num_context_features - index);
      }
      ny::trie t;
      pecco::build_trie (&t, "dict trie", dict, str, len, val, _opt.verbose > 0, "ab");
      for (sbag_t::iterator it = sbag.begin (); it != sbag.end (); ++it)
        delete [] it->first;
    }
    _dict = new dict_t (dict.c_str ());
    if (_opt.verbose > 0)
      std::fprintf (stderr, "done. (# strings + 1 (unseen) = %d).\n",
                    _dict->num_lexical_features ());
    _context_feature_bits.resize (_dict->context_feature_bit_len (), 0);
  }
  void parser::_setup_learner () {
    switch (_opt.learner) { // they should be newed earlier
#ifdef USE_OPAL
      case OPAL:
        _opal_opt.set (_opt.learner_argc, _opt.learner_argv);
        _opal = new opal::Model (_opal_opt);
        break;
#endif
#ifdef USE_SVM
      case SVM:
        _tiny_param.set (_opt.learner_argc, _opt.learner_argv);
        if (_tiny_param.kernel_type != TinySVM::POLY || _tiny_param.degree > 4)
          ny::print_err ("only polynomial kernel [-t 1] of [-d] <= 4 is supported in SVM.\n");
        break;
#endif
#ifdef USE_MAXENT
      case MAXENT:
        _maxent_opt.set (_opt.learner_argc, _opt.learner_argv);
        _libme = new ME_Model;
        break;
#endif
      default: break;
    }
  }
  void parser::_cleanup_learner () {
    switch (_opt.learner) {
#ifdef USE_OPAL
      case OPAL:   delete _opal;    break;
#endif
#ifdef USE_SVM
      case SVM:    delete _tinysvm; break;
#endif
#ifdef USE_MAXENT
      case MAXENT: delete _libme;   break;
#endif
      default: break;
    }
  }
  void parser::_switch_classifier (const input_t in) // exchange classifiers
  { _pecco = in == CHUNK ? _pecco_chunk : _pecco_depnd; }
  void parser::_setup_classifier (const input_t in, int argc, char ** argv) {
    std::string model (_opt.model_dir + "/" + input0[in]);
    if (in == DEPND) {
      char sigparse[16]; std::sprintf (sigparse, ".p%d", _opt.parser);
      model += sigparse;
    }
    if (_opt.mode == BOTH) { // induce an appropriate classifier from input model
#ifdef USE_OPAL
      if (_opt.learner == OPAL && _opal_opt.kernel == opal::POLY)
        _opt.learner = SVM; // fake
#endif
    } else {
      FILE * fp = std::fopen (model.c_str (), "r");
      if (! fp || std::feof (fp))
        ny::print_err (HERE "no model found: %s; train a model first [-t 0]\n",
                       model.c_str ());
      switch (std::fgetc (fp)) {
        case  0 :
        case '#': _opt.learner = OPAL;   break;
        case 'o': // delegate
        case 'T': _opt.learner = SVM;    break;
        case '-':
        case '+': _opt.learner = MAXENT; break;
        default:  ny::print_err (HERE "unknown model type found: %s;");
      }
#ifndef USE_OPAL
      if (_opt.learner == OPAL)
        ny::print_err (HERE "unsupported model found, try jdepp_pa\n");
#endif
      std::fclose (fp);
    }
    if  (_opt.learner == OPAL) { // opal as a classifier
#ifdef USE_OPAL
      opal::option opal_opt (argc, argv);
      opal_opt.model = model.c_str ();
      _pecco = new pecco::pecco (opal_opt, reinterpret_cast <opal::Model *> (0));
      _pecco->load (opal_opt.model);
#endif
    } else {
      const std::string train (model + ".train");
      const std::string event (model + ".event");
      _pecco_opt.set (argc, argv);
      _pecco_opt.model = model.c_str ();
      _pecco_opt.train = train.c_str ();
      _pecco_opt.event = event;
      if (_opt.learner == SVM) {
#if defined (USE_SVM) || defined (USE_OPAL)
        _pecco_opt.event += std::string (".s") + _pecco_opt.sigma; // approx
        _pecco = new pecco::pecco (_pecco_opt,
                                   static_cast <pecco::kernel_model *> (0));
#endif
      } else {
#ifdef USE_MAXENT
        _pecco = new pecco::pecco (_pecco_opt,
                                   static_cast <pecco::linear_model *> (0));
#endif
      }
      _pecco->load (_pecco_opt.model);
    }
    if (in == CHUNK)  _pecco_chunk = _pecco; else _pecco_depnd = _pecco;
  }
  void parser::_cleanup_classifier (const input_t in) {
    _switch_classifier (in);
    _pecco->printStat ();
    delete _pecco;
  }
#ifdef USE_MAXENT
  void parser::_project (ME_Sample &ms) const {
    char key[64];
    for (ny::fv_it beg (_fv.begin ()), end (_fv.end ()), it (beg);
         it != end; ++it) {
      const int pos_i = std::sprintf (key, "%d", *it);
      ms.add_feature (key);
      if (_maxent_opt.degree != 1)
        for (ny::fv_it jt = beg; jt != it; ++jt) {
          const int pos_j = pos_i + std::sprintf (key + pos_i, ":%d", *jt);
          ms.add_feature (key);
          if (_maxent_opt.degree != 2)
            for (ny::fv_it kt = beg; kt != jt; ++kt) {
              std::sprintf (key + pos_j, ":%d", *kt);
              ms.add_feature (key);
            }
        }
    }
  }
#endif
  void parser::run () {
#ifdef USE_TIMER
    if (_opt.verbose > 0)
      std::fprintf (stderr, "Processor Speed: %.3LfGHz.\n",
                    ny::Timer::clock / 1000);
#endif
    if (_opt.input == RAW && _opt.mode != PARSE)
#ifdef USE_AS_STANDALONE
      ny::print_err ("You can input RAW sentences [-I 0] only for parsing [-t 1].\n");
#else
      ny::print_err ("You can input POS-tagged sentences [-I 0] only for parsing [-t 1].\n");
#endif

    TIMER (_dict_t->startTimer ());
    _setMorphDic ();
    TIMER (_dict_t->stopTimer ());
    // learn
    if  (_opt.mode == LEARN || _opt.mode == BOTH) { // learn or both
      _setup_learner ();
      _analyze <LEARN> ();
      _learn ();
      _cleanup_learner ();
    }
    // parse, cache
    if (_opt.mode != LEARN) {
      if (_opt.input != DEPND)
        _setup_classifier (CHUNK, _opt.chunk_argc, _opt.chunk_argv);
      if (_opt.input != CHUNK)
        _setup_classifier (DEPND, _opt.depnd_argc, _opt.depnd_argv);
      if (_opt.mode == CACHE) {
        if (_opt.learner == OPAL)
          ny::print_err ("needless to cache in opal classifier [-t 0].\n");
        _analyze <CACHE> ();
      } else {
        _analyze <PARSE> ();
        if (_opt.input == CHUNK) _chunk_stat.print ();
        if (_opt.input == DEPND) _depnd_stat.print ();
      }
      // cleanup
      TIMER (_timer_pool.print ());
      if (_opt.input != DEPND) _cleanup_classifier (CHUNK);
      if (_opt.input != CHUNK) _cleanup_classifier (DEPND);
    }
  }
}
// ToDo
// - output model info, or at least add signature
// - implement POS tagger / word segmenter faster than MeCab
