// J.DepP -- Japanese Dependency Parsers
//  $Id: pdep.h 762 2012-01-04 15:37:59Z ynaga $
// Copyright (c) 2008-2011 Naoki Yoshinaga <ynaga@tkl.iis.u-tokyo.ac.jp>
#ifndef PDEP_H
#define PDEP_H

#include <unistd.h>
#include <fcntl.h>
#include <cstdio>
#include <cmath>
#include <cassert>
#include <set>
#include <map>
#include <stack>
#include <list>
#ifdef USE_HASH
#include <tr1/unordered_map>
#endif
#include "typedef.h"
#include "jdepp_conf.h"
// trie type
#if defined (USE_DARTS) || defined (USE_DARTS_CLONE)
#include <darts.h>
#endif
#ifdef USE_CEDAR
#include "cedar.h"
#endif
#ifdef USE_TIMER
#include "timer.h"
#endif
#if defined (USE_OPAL) || defined (USE_SVM)
#include "kernel.h"
#endif
#ifdef USE_SVM
#include <tinysvm.h>
#endif
#ifdef USE_OPAL
#define USE_PRUNE
#define USE_PMT
#include "opal_conf.h"
#include "pa.h"
#endif
#ifdef USE_MAXENT
#include "maxent.h"
#include "linear.h"
#endif
#include "pecco.h"
#ifdef USE_AS_STANDALONE
#include <mecab.h>
#endif
#define JDEPP_COPYRIGHT  "J.DepP - Japanese Dependency Parser\n\
Copyright (c) 2008-2011 Naoki Yoshinaga\n\
\n\
Usage: %s [options] -- [learner options] -- [chunker classifier options] -- [parser classifier options] < test\n\
\n\
test    test file             \n\
\n"

#define JDEPP_OPT0 "Optional parameters in training / testing:\n\
  -t, --type=TYPE             select running mode of J.DepP\n\
                                0 - learn\n\
                              * 1 - parse\n\
                                2 - both\n\
                                3 - cache\n\
  -c, --corpus=FILE           training corpus in JDEPP format ('train.JDP')\n\
  -m, --model-dir=DIR         model directory ('model')\n\
  -p, --parser=TYPE           select parsing algorithm\n\
                              * 0 - shift reduce\n\
                                1 - cascaded chunking\n\
                                2 - backward\n\
                                3 - tournament\n"

#ifdef USE_AS_STANDALONE
#define JDEPP_OPT1 "  -I, --input-format=TYPE     select type of input format\n\
                              * 0 - RAW sentences\n\
                                1 - + POS / BUNSETSU annotation\n\
                                2 - + DEPENDENCY annotation\n\
\n"
#else
#define JDEPP_OPT1 "  -I, --input-format=TYPE     select type of input format\n\
                              * 0 - POS-tagged sentences\n\
                                1 - + BUNSETSU annotation\n\
                                2 - + DEPENDENCY annotation\n\
\n"
#endif

#if defined (USE_OPAL) && defined (USE_SVM) && defined (USE_MAXENT)
#define JDEPP_OPT_TRAIN "Optional parameters in training:\n\
  -l, --learner=TYPE          select type of learning library\n\
                              * 0 - OPAL\n\
                                1 - SVM\n\
                                2 - MaxEnt\n\
  -n, --max-sent=INT          max. # processing sentences (0: all)\n\
\n"
#else
#define JDEPP_OPT_TRAIN "Optional parameters in training:\n\
  -n, --max-sent=INT          max. # processing sentences (0: all)\n\
\n"
#endif

#ifdef USE_AS_STANDALONE
#define JDEPP_OPT_TEST "Optional parameters in testing:\n\
  -d, --mecab-dic=DIR         use MeCab dict in DIR for POS tagging\n\
\n"
#else
#define JDEPP_OPT_TEST
#endif

#define JDEPP_OPT_MISC "Misc.:\n\
  -v, --verbose=INT           verbosity level (1)\n\
  -h, --help                  show this help and exit\n"

#define MAXENT_OPT "\nOptions for MaxEnt learners are as follows:\n\
  -d, --degree=INT            maximum degree of feature combinations (<=3)\n\
  -l, --algorithm=INT         select type of optimization algorithm\n\
                              * 0 - SGD-L1\n\
                                1 - OWLQN-L1\n\
                                2 - LBFGS-L2\n\
  -c, --reg-cost=INT          cost of regularization\n\
\n"

#if defined (USE_OPAL) && defined (USE_SVM) && defined (USE_MAXENT)
#ifdef USE_AS_STANDALONE
static const  char * jdepp_short_options = "t:c:m:p:I:b:l:n:d:v:h";
#else
static const  char * jdepp_short_options = "t:c:m:p:I:b:l:n:v:h";
#endif
#else
#ifdef USE_AS_STANDALONE
static const  char * jdepp_short_options = "t:c:m:p:I:b:n:d:v:h";
#else
static const  char * jdepp_short_options = "t:c:m:p:I:b:n:v:h";
#endif
#endif
static struct option jdepp_long_options[] = {
  {"type",         required_argument, NULL, 't'},
  {"corpus",       required_argument, NULL, 'c'},
  {"model-dir",    required_argument, NULL, 'm'},
  {"parser",       required_argument, NULL, 'p'},
  {"input-format", required_argument, NULL, 'I'},
  {"cluster-bits", required_argument, NULL, 'b'},
#if defined (USE_OPAL) && defined (USE_SVM) && defined (USE_MAXENT)
  {"learner",      required_argument, NULL, 'l'},
#endif
  {"max-sent",     required_argument, NULL, 'n'},
#ifdef USE_AS_STANDALONE
  {"mecab-dic",    required_argument, NULL, 'd'},
#endif
  {"verbose",      required_argument, NULL, 'v'},
  {"help",         no_argument,       NULL, 'h'},
  {NULL, 0, NULL, 0}
};

static const  char * maxent_short_options = "d:l:c:h";
static struct option maxent_long_options[] = {
  {"degree",       required_argument, NULL, 'd'},
  {"algorithm",    required_argument, NULL, 'l'},
  {"reg-cost",     required_argument, NULL, 'c'},
  {"help",         no_argument,       NULL, 'h'},
  {NULL, 0, NULL, 0}
};

extern char * optarg;
extern int    optind;

namespace pdep {
  // type alias
  typedef std::vector <uint64_t> flag_t;
  static const size_t FLAG_LEN = sizeof (flag_t::value_type) * 8;
#if   defined (USE_JUMAN_POS)
#if   defined (USE_MECAB)
  enum field_t {SURF, POS1, POS2, TYPE, INFL, FIN, YOMI, NUM_FIELD};
#elif defined (USE_JUMAN)
  enum field_t {SURF, YOMI, FIN, POS1, POS_ID1, POS2, POS_ID2, TYPE, TYPE_ID, INFL, INFL_ID, NUM_FIELD};
#endif
#elif defined (USE_IPA_POS)
  enum field_t {SURF, POS1, POS2, POS3, POS4, TYPE, INFL, FIN, YOMI, PRON, NUM_FIELD};
#endif
  // static const variables
  template <typename T>
  static void widen (T * &array, const int &avail, const int &filled = 0) {
    T * tmp = static_cast <T*> (::operator new (avail * sizeof (T)));
    if (filled) {
      std::memcpy (&tmp[0], &array[0], sizeof (T) * filled);   // delegate
      for (int i = filled; i < avail; ++i) new (&tmp[i]) T (); // fill
    }
    std::swap (tmp, array);
    if (tmp) ::operator delete (tmp); // don't destruct delegated resource
  }
  template <typename T>
  T strton (const char * s, char ** error)
  { return static_cast <T> (std::strtol (s, error, 10)); }
  //
  enum process_t     { LEARN, PARSE, BOTH, CACHE };
  enum parser_t      { LINEAR, CHUNKING, BACKWARD, TOURNAMENT };
  enum learner_t     { OPAL, SVM, MAXENT };
  enum input_t       { RAW, CHUNK, DEPND };
  enum maxent_algo_t { SGD, OWLQN, LBFGS }; // MaxEnt optimizer
  // The command-line arguments will override the following default parameters
  // default learner parameters
  static const size_t JDEPP_OPAL_ARGC_CHUNK = 9;
  static const char * JDEPP_OPAL_ARGV_CHUNK[] =
    {"--", "-t", "1", "-d", "2", "-c", "0.001",   "-i", "40"};
  static const size_t JDEPP_OPAL_ARGC_DEPND = 9;
  static const char * JDEPP_OPAL_ARGV_DEPND[] =
    {"--", "-t", "1", "-d", "2", "-c", "0.00005", "-i", "40"};
  // default classifier parameters
  static const size_t JDEPP_PECCO_ARGC_CHUNK = 3;
  static const char * JDEPP_PECCO_ARGV_CHUNK[] = {"--", "-s", "0.03"};
  static const size_t JDEPP_PECCO_ARGC_DEPND = 3;
  static const char * JDEPP_PECCO_ARGV_DEPND[] = {"--", "-s", "0.02"};
  // {"--", "-t", "1", "-s", "0.02", "-i", "8"} # faster
  // For those who wants an extremely efficient dependency parser
  // default learner parameters
  // static const size_t JDEPP_OPAL_ARGC_CHUNK = 7;
  // static const char * JDEPP_OPAL_ARGV_CHUNK[] =
  //   {"--", "-t", "0", "-c", "0.05",  "-i", "40"};
  // static const size_t JDEPP_OPAL_ARGC_DEPND = 7;
  // static const char * JDEPP_OPAL_ARGV_DEPND =
  //   {"--", "-t", "0", "-c", "0.001", "-i", "40"};
  // // default classifier parameters
  // static const size_t JDEPP_PECCO_ARGC_CHUNK = 1;
  // static const char * JDEPP_PECCO_ARGV_CHUNK[] = {"--"};
  // static const size_t JDEPP_PECCO_ARGC_DEPND = 1;
  // static const char * JDEPP_PECCO_ARGV_DEPND[] = {"--"};
  class option { // option handler
  public:
    const char *com, *train;
    std::string model_dir;
    //
    process_t mode;
    parser_t  parser;
    ny::uint  cbits;
    ny::uint  clen;
    learner_t learner;
    ny::uint  max_sent;
    input_t   input;
#ifdef USE_AS_STANDALONE
    char *    mecab_dic;
#endif
    int       verbose;
    int       learner_argc;
    char **   learner_argv;
    int       depnd_argc;
    char **   depnd_argv;
    int       chunk_argc;
    char **   chunk_argv;
    //
    option (int argc, char ** argv) :
      com (argv[0]), train ("train.JDP"),
#ifdef USE_STACKING
      model_dir ("model_stack"),
#else
      model_dir ("model"),
#endif
      mode (PARSE), parser (LINEAR), cbits (0), clen (0),
#if   defined (USE_OPAL)
      learner (OPAL),
#elif defined (USE_SVM)
      learner (SVM),
#elif defined (USE_MAXENT)
      learner (MAXENT),
#endif
      max_sent (0), input (RAW),
#ifdef USE_AS_STANDALONE
#ifdef JDEPP_MECAB_DIC
      mecab_dic (JDEPP_MECAB_DIC),
#else
      mecab_dic (0),
#endif
#endif
      verbose (0), learner_argc (0), learner_argv (0), depnd_argc (0), depnd_argv (0), chunk_argc (0), chunk_argv (0) {
      // getOpt
      if (argc == 0) return;
      optind = 1;
      while (1) {
        int opt = getopt_long (argc, argv,
                               jdepp_short_options, jdepp_long_options, NULL);
        if (opt == -1) break;
        char * err = NULL;
        switch (opt) {
          case 't': mode      = strton <process_t> (optarg, &err); break;
          case 'c': train     = optarg; break;
          case 'm': model_dir = optarg; break;
          case 'p': parser    = strton <parser_t>  (optarg, &err); break;
          case 'I': input     = strton <input_t>   (optarg, &err); break;
          case 'b':
            do {
              const ny::uint depth = std::strtol (optarg, &optarg, 10);
              cbits |= 1 << (depth - 1);
              clen = std::max (clen, depth);
            } while (*optarg++ != '\0');
            break;
            // training parameters
#if defined (USE_OPAL) && defined (USE_SVM) && defined (USE_MAXENT)
          case 'l': learner   = strton <learner_t> (optarg, &err); break;
#endif
          case 'n': max_sent  = std::strtol (optarg, &err, 10);    break;
            // misc
#ifdef USE_AS_STANDALONE
          case 'd': mecab_dic = optarg; break;
#endif
          case 'v': verbose   = std::strtol (optarg, &err, 10);    break;
          case 'h': printCredit (); printHelp (); std::exit (0);   break;
          default:  printCredit (); std::exit (0);
        }
        if (err && *err)
          ny::print_err (HERE "unrecognized option value: %s\n", optarg);
      }
      // errors & warnings
      if (learner != OPAL && learner != SVM && learner != MAXENT)
        ny::print_err (HERE "unknown learner [-l].\n");
      if (mode != LEARN && mode != PARSE && mode != BOTH && mode != CACHE)
        ny::print_err (HERE "unknown running mode [-t].\n");
      if (parser != LINEAR && parser != CHUNKING && parser != BACKWARD &&
          parser != TOURNAMENT)
        ny::print_err (HERE "unknown parsing algorithm [-p].\n");
      if (input != RAW && input != CHUNK && input != DEPND)
        ny::print_err (HERE "unknown input format [-I].\n");
      struct stat st;
      if (stat (model_dir.c_str (), &st) != 0)
        ny::print_err (HERE "no such directory: %s [-m]\n", model_dir.c_str ());
#ifdef USE_AS_STANDALONE
      if (mecab_dic && stat (mecab_dic, &st) != 0)
          ny::print_err (HERE "no such file or directory: %s", mecab_dic);
#endif
      if (input == CHUNK && parser != LINEAR)
        std::fprintf (stderr, HERE "parsing algorithm [-p] is ignored in training a chunker.\n");
      // learner options
      if (std::strcmp (argv[optind - 1], "--") == 0) --optind;
      _set_library_options (optind, argc, argv, learner_argc, learner_argv);
      // classifier options for bunsetsu chunker
      _set_library_options (optind, argc, argv, chunk_argc, chunk_argv);
      // classifier options for dependency parser
      _set_library_options (optind, argc, argv, depnd_argc, depnd_argv);
      // default parameters
      if (learner == OPAL)
        if (learner_argc <= 1 && chunk_argc <= 1 && depnd_argc <= 1) {
          if (input == CHUNK)
            learner_argc = JDEPP_OPAL_ARGC_CHUNK,
            learner_argv = const_cast <char **> (&JDEPP_OPAL_ARGV_CHUNK[0]);
          else if (input == DEPND)
            learner_argc = JDEPP_OPAL_ARGC_DEPND,
            learner_argv = const_cast <char **> (&JDEPP_OPAL_ARGV_DEPND[0]);
          chunk_argc = JDEPP_PECCO_ARGC_CHUNK,
          chunk_argv = const_cast <char **> (&JDEPP_PECCO_ARGV_CHUNK[0]);
          depnd_argc = JDEPP_PECCO_ARGC_DEPND,
          depnd_argv = const_cast <char **> (&JDEPP_PECCO_ARGV_DEPND[0]);
        }
    }
    void printCredit () { std::fprintf (stderr, JDEPP_COPYRIGHT, com); }
    void printHelp   () { std::fprintf (stderr, JDEPP_OPT0 JDEPP_OPT1 JDEPP_OPT_TRAIN JDEPP_OPT_TEST JDEPP_OPT_MISC); }
  private:
    void _set_library_options (int &i, const int argc, char ** argv,
                               int &libargc, char ** &libargv) {
      if (i < argc) {
        if (std::strcmp (argv[optind], "--") == 0) { // library options
          libargv = &argv[optind];
          libargc = 1;
          while (optind + libargc < argc &&
                 std::strcmp (libargv[libargc], "--") != 0)
            ++libargc;
          i += libargc;
        } else {
          printCredit ();
          ny::print_err ("Type `%s --help' for option details.\n", com);
        }
      }
    }
  };
  struct maxent_option { // option handler
    maxent_algo_t algo;
    ny::uint      degree;
    ny::fl_t      reg_cost;
    //
    maxent_option () : algo (SGD), degree (2), reg_cost (1.0) {}
    maxent_option (int argc, char ** argv) : algo (SGD), degree (2), reg_cost (1.0) {
      set (argc, argv);
    }
    void set (int argc, char ** argv) { // getOpt
      if (argc == 0) return;
      optind = 1;
      while (1) {
        int opt = getopt_long (argc, argv,
                               maxent_short_options, maxent_long_options, NULL);
        if (opt == -1) break;
        char * err = NULL;
        switch (opt) {
          case 'd': degree   = std::strtol (optarg, &err, 10); break;
          case 'l': algo     = strton <maxent_algo_t> (optarg, &err); break;
          case 'c': reg_cost = std::strtod (optarg, &err); break;
          case 'h': printHelp (); std::exit (0); break;
          default:  std::exit (0);
        }
        if (err && *err)
          ny::print_err ("unrecognized option value: %s\n", optarg);
      }
      // errors
      if (algo != SGD && algo != OWLQN && algo != LBFGS)
        ny::print_err (HERE "unknown optimization algorithm [-l]");
      if (degree == 0 || degree >= 4)
        ny::print_err (HERE "set degree <= 3 for [-d].\n");

    }
    void printHelp () { std::fprintf (stderr, MAXENT_OPT); }
  };
  // dictionary utility variables
  typedef std::map <const char*, ny::uint, ny::pless <char> > sbag_t;
  class dict_base_t : private ny::Uncopyable {
  private:
    char *    _data_ptr;
    ny::trie  _data;
    ny::uint  _num_lexical_features;
    ny::uint  _num_context_features;
  protected:
    dict_base_t (const char *fn) : _data_ptr (0), _data (), _num_lexical_features (0), _num_context_features (0) {
      FILE * fp = std::fopen (fn, "rb");
      if (std::fread (&_num_lexical_features, sizeof (ny::uint), 1, fp) != 1 ||
          std::fread (&_num_context_features, sizeof (ny::uint), 1, fp) != 1)
        ny::print_err (HERE "broken dic: delete %s\n", fn);
      ++_num_lexical_features; // reserve room for unseen feature
      size_t offset = std::ftell (fp);
      std::fseek (fp, 0,      SEEK_END);
      size_t size   = std::ftell (fp) - offset;
      std::fseek (fp, offset, SEEK_SET);
      _data_ptr = new char[size];
      std::fread (_data_ptr, sizeof (char), size, fp);
      _data.set_array (_data_ptr, size / _data.unit_size ());
      std::fclose (fp);
    }
    ~dict_base_t () { delete [] _data_ptr; }
  public:
    int      lookup (const char *key, size_t len = 0) const
    { return _data.exactMatchSearch <ny::trie::result_type> (key, len); }
    size_t   context_feature_bit_len () const
    { return (_num_context_features - 1) / FLAG_LEN + 1; }
    bool     is_context_feature (const ny::uint id) const
    { return id < _num_context_features; }
    ny::uint num_lexical_features () const { return _num_lexical_features; }
  };
  class dict_t : public dict_base_t {
  public: // surface id aliases
    const int  comma;
    const int  period;
    const int  particle;
    const int  bracket_start;
    const int  bracket_end;
    const int  special;
#if defined (USE_JUMAN_POS) and defined (USE_EMNLP_FEAT)
    const int  suffix;
#endif
    const int  unseen;
    dict_t (const char *fn) :
      dict_base_t (fn), comma (lookup (COMMA)), period (lookup (PERIOD)), particle (lookup (POST_PARTICLE)), bracket_start (lookup (BRACKET_START)), bracket_end (lookup (BRACKET_END)), special (lookup (SPECIAL)),
#if defined (USE_JUMAN_POS) and defined (USE_EMNLP_FEAT)
      suffix (lookup (SUFFIX)),
#endif
      unseen (num_lexical_features () - 1) {}
  };
  class Morph {
  private:
    int    _field[NUM_FIELD];
  public:
    const char * surface;
    size_t length;
    const char * feature;
    Morph () : _field (), surface (0), length (0), feature (0) {}
    Morph (const Morph& m) : _field (), surface (m.surface), length (m.length), feature (m.feature)
    { std::copy (&m._field[0], &m._field[0] + NUM_FIELD, &_field[0]); }
    ~Morph () {}
    void set (char * p, const ssize_t len, const dict_t * dict) {
      surface = p;
      char * const p_end = p + len;
      *p_end = '\0'; while (p != p_end && *p != SURFACE_END) ++p; // juman
      length = p - surface; feature = ++p;
      set (dict);
    }
    void set (const dict_t * dict) {
      int n = dict->lookup (surface, length); // read surface
      _field[SURF] = (n >= 0) ? n : dict->unseen;
      // read feature
      ny::uint i = 1;
      for (const char * p (feature), * f (p); i < NUM_FIELD; f = p, ++i) {
        while (*p != '\0' && *p != FEATURE_SEP) ++p;
        if (*p) ++p; else break;
        // if (i <= POS2 || i == INFL) {
        if (i == SURF || i == POS1 || i == POS2 || i == INFL) {
          int n = dict->lookup (f, p - f - 1);
          _field[i] = (n >= 0) ? n : dict->unseen;
        }
      }
#ifdef USE_JUMAN_POS
      if (i < NUM_FIELD) {
        std::fwrite (surface, sizeof (char), length, stderr);
        ny::print_err (HERE "# fields, %d, is less than %d.\n", i, NUM_FIELD);
      }
#endif
    }
    int surf () const { return _field[SURF]; }
    int pos1 () const { return _field[POS1]; }
    int pos2 () const { return _field[POS2]; }
    int infl () const { return _field[INFL]; }
    int cluster (int i) const { return _field[NUM_FIELD + i]; }
  };
  class Bunsetsu {
  private:
    void _set_fbits (ny::uint i) {
      fbits[i / FLAG_LEN]
        |= (static_cast <flag_t::value_type> (1) << (i % FLAG_LEN));
    }
  public:
    int     mpos;     // start position
    int     mlen;     // number of morphs
    flag_t  fbits;
    int     head;     // head_morph offset
    int     tail;     // tail
    int     id;       // bunsetsu id
    int     did;      // dest id (gold)
    int     est_did;  // dest id (estimated)
    int     cand;
    char    dt;
    char    candt;
    int     bracket;  // has bracket
    bool    comma;    // has comma
    bool    period;   // has period
    // char dtype;
    Bunsetsu () : mpos (0), mlen (0), fbits (), head (-1), tail (-1), id (-1), did (-1), est_did (-1), cand (-1), dt ('D'), candt ('D'), bracket (0), comma (false), period (false) {}
    Bunsetsu (const Bunsetsu &b) : mpos (b.mpos), mlen (b.mlen), fbits (b.fbits), head (b.head), tail (b.tail), id (b.id), did (b.did), est_did (b.est_did), cand (b.cand), dt (b.dt), candt (b.candt), bracket (b.bracket), comma (b.comma), period (b.period) {}
    ~Bunsetsu () {}
    void clear () { // leave other members overridden
      std::fill (fbits.begin (), fbits.end (), 0);
      head = tail = est_did = cand = -1; dt = 'D'; candt = 'D';
      mpos = mlen = bracket = 0;
      comma = period = false;
    }
    void set (char * p, const ssize_t len) { // ex. '* 1 7D'
#ifdef USE_STACKING 
      const char * p_end (p + len);
      id  = ny::strton <int> (p + 2, &p);
      did = ny::strton <int> (++p, &p);
      if (++p != p_end) {
        cand  = ny::strton <int> (++p, &p); // ignored in testing
        candt = *p;
      }
#else
      id  = ny::strton <int> (p + 2, &p);
      did = ny::strton <int> (++p, &p);
      dt  = *p;
#endif
    }
    void set (const int id_) {
      id  = id_;
      did = -1;
#ifdef USE_STACKING
      cand = -1;
#endif
    }
    bool setup (const dict_t * dict, const Morph * morph) {
      fbits.resize (dict->context_feature_bit_len (), 0);
      for (int i = mlen - 1; i >= 0; --i) {
        const Morph &m = morph[i];
        if (m.pos1 () == dict->special) {
          if      (m.pos2 () == dict->comma)  comma  |= true;
          else if (m.pos2 () == dict->period) period |= true;
          else {
#ifdef USE_EMNLP_FEAT
            if (dict->is_context_feature (m.surf ()))
              _set_fbits (m.surf ());
#endif
            if      (m.pos2 () == dict->bracket_end)   --bracket;
            else if (m.pos2 () == dict->bracket_start) ++bracket;
          }
        } else {
          if (tail == -1) {
            tail = i;
#ifdef USE_EMNLP_FEAT
            if (m.pos1 () == dict->particle)
              if (dict->is_context_feature (m.surf ()))
                _set_fbits (m.surf ());
#endif
          }
#ifdef USE_EMNLP_FEAT
          // base ; 0.9058; 0.5206
          if (head == -1 &&
              m.pos1 () != dict->particle && m.pos1 () != dict->suffix)
              // m.pos1 () != dict->particle) // 0.9069; 0.5249
            head = i;
#else
          if (m.pos1 () == dict->particle) {
            if (dict->is_context_feature (m.surf ()))
              _set_fbits (m.surf ());
          } else {
            if (head == -1) head = i;
          }
#endif
        }
      }
      if (did < 0 || did > id)  return true;
#ifndef NDEBUG
      ny::print_err ("\tbroken dependency?: %d->%d\n", id, did);
#endif
      return false;
    }
  };
  class Sentence : private ny::Uncopyable {
  private:
    int        _bavail;
    int        _mavail;
    char       _res[IOBUF_SIZE]; // save output
    char*      _ptr;             // current position in result buffer
  public:
    Bunsetsu * bun;
    Morph    * morph;
    int        blen;
    int        mlen;
    Sentence () : _bavail (1), _mavail (1), _res (), _ptr (&_res[0]), bun (static_cast <Bunsetsu *> (::operator new (_bavail * sizeof (Bunsetsu)))), morph (static_cast <Morph *> (::operator new (_mavail * sizeof (Morph)))), blen (0), mlen (0) {
      for (int i = 0; i < _bavail; ++i) new (&bun[i])   Bunsetsu ();
      for (int i = 0; i < _mavail; ++i) new (&morph[i]) Morph    ();
    }
    ~Sentence () {
      for (int i = 0; i < _bavail; ++i) bun[i].~Bunsetsu ();
      ::operator delete (bun);
      for (int i = 0; i < _mavail; ++i) morph[i].~Morph ();
      ::operator delete (morph);
    }
    void clear () {
      if (blen) do { bun[--blen].clear (); } while (blen);
      mlen = 0;
    };
    void setHeader (char * cs, const ssize_t len) // skip comment
    { if (_ptr == &_res[0]) { std::memcpy (_ptr, cs, len); _ptr += len; } }
    void setup (const dict_t *dict) { // complete information
      for (int i = 0; i < blen; ++i) {
        bun[i].mlen = (i == blen - 1 ? mlen : bun[i+1].mpos) - bun[i].mpos;
        bun[i].setup (dict, &morph[bun[i].mpos]);
      }
    }
    void addBunsetsu (char * cs, const size_t len, const int mpos) {
      if (blen == _bavail) { _bavail <<= 1; widen (bun, _bavail, blen); }
      bun[blen].mpos = mpos;
      bun[blen].set (cs, len);
      ++blen;
    }
    void addBunsetsu (const int mpos) {
      if (blen == _bavail) { _bavail <<= 1; widen (bun, _bavail, blen); }
      bun[blen].mpos = mpos;
      bun[blen].set (blen);
      ++blen;
    }
    void addMorph (char * cs, const ssize_t len, const dict_t * dict) {
      if (mlen == _mavail) { _mavail <<= 1; widen (morph, _mavail, mlen); }
      morph[mlen].set (cs, len, dict);
      ++mlen;
    }
    void addMorph (const char * surface, const size_t length,
                   const char * feature, const dict_t * dict) {
      if (mlen == _mavail) { _mavail <<= 1; widen (morph, _mavail, mlen); }
      morph[mlen].surface = surface;
      morph[mlen].length  = length;
      morph[mlen].feature = feature;
      morph[mlen].set (dict);
      ++mlen;
    }
    void print () {
      for (int i = 0; i < blen; ++i) {
        const Bunsetsu &b = bun[i];
        const Morph *  ms = &morph[b.mpos];
#ifdef USE_STACKING
        _ptr += std::sprintf (_ptr, "* %d %dD %d%c\n", i, b.est_did, b.cand, b.candt);
#else
        _ptr
          += b.did >= 0 ?
          std::sprintf (_ptr, "* %d %dD#%d%c\n", i, b.est_did, b.did, b.dt) :
          std::sprintf (_ptr, "* %d %dD\n", i, b.est_did);
#endif
        for (int j = 0; j < b.mlen; ++j) {
          const Morph &m = ms[j];
          std::memcpy (_ptr, m.surface, m.length); _ptr += m.length;
          _ptr += std::sprintf (_ptr, "%c%s\n", SURFACE_END, m.feature);
        }
      }
      std::memcpy (_ptr, "EOS\n", 4); _ptr += 4;
      char * check = &_res[0];
      while (check < _ptr) check += write (1, check, _ptr - check);
      _ptr = &_res[0];
    }
  };
  template <typename T>
  struct stat_base {
    ny::uint snum;  // # sentence
    ny::uint scorr; // # sentence correctly recognized
    stat_base () : snum (0), scorr (0) {}
    void print () {
      static_cast <T*> (this)->print_impl ();
      std::fprintf (stderr, "acc. (complete)\t%.4f (%5d/%5d)\n\n",
                    scorr * 1.0 / snum, scorr, snum);
    }
  protected:
    ~stat_base () {}
  };
  struct chunk_stat : public stat_base <chunk_stat> {
    ny::uint pp;  // # chunks correctly recognized
    ny::uint np;  // # chunks incorrectly recognized
    ny::uint pn;  // # chunks missed
    chunk_stat () : pp (0), np (0), pn (0) {}
    void print_impl () {
      std::fprintf (stderr, "J.DepP performance statistics (chunk):\n");
      const ny::fl_t prec (pp * 1.0 / (pp + np)), rec (pp * 1.0 / (pp + pn));
      std::fprintf (stderr, "precision\t%.4f (%5d/%5d)\n", prec, pp, pp + np);
      std::fprintf (stderr, "recall   \t%.4f (%5d/%5d)\n", rec,  pp, pp + pn);
      std::fprintf (stderr, "f1       \t%.4f\n", 2 * prec * rec / (prec + rec));
    }
  };
  struct depnd_stat : public stat_base <depnd_stat> {
    ny::uint bnum;   // # dependencies
    ny::uint bcorr;  // # dependencies correctly recognized
    depnd_stat () : bnum (0), bcorr (0) {}
    void print_impl () {
      std::fprintf (stderr, "J.DepP performance statistics (depnd):\n");
      std::fprintf (stderr, "acc. (partial)\t%.4f (%5d/%5d)\n",
                    bcorr * 1.0 / bnum, bcorr, bnum);
    }
  };
  struct chunk_info {
    int  id;
    chunk_info (int id_) : id (id_) {};
    // bool done; // deterministic
    // chunk_info (int id_, bool done_) : id (id_), done (done_) {};
  };
  class parser : private ny::Uncopyable {
  private:
    option           _opt;
    pecco::option    _pecco_opt;
    pecco::pecco *   _pecco;
    pecco::pecco *   _pecco_chunk;
    pecco::pecco *   _pecco_depnd;
#ifdef USE_OPAL
    opal::option     _opal_opt;
    opal::Model *    _opal;
    opal::mem_pool <opal::ex_t> _ex_pool;
#endif
#ifdef USE_SVM
    TinySVM::Param   _tiny_param;
    TinySVM::Model * _tinysvm;
#endif
#ifdef USE_MAXENT
    maxent_option    _maxent_opt;
    ME_Model *       _libme;
#endif
    // variables to handle events
    Sentence *       _s;
    dict_t *         _dict;
    ny::uint         _fi; // offset of feature index
    ny::fv_t         _fv; // feature vector of the example
    flag_t           _context_feature_bits;
    chunk_stat       _chunk_stat;
    depnd_stat       _depnd_stat;
    FILE *           _writer;
    // timer
#ifdef USE_TIMER
    ny::TimerPool    _timer_pool;
    ny::Timer *      _io_t;
    ny::Timer *      _dict_t;
    ny::Timer *      _preproc_t;
    ny::Timer *      _chunk_t;
    ny::Timer *      _depnd_t;
    ny::Timer *      _classify_t;
#endif
    void _print_ex (const bool flag) const {
      std::fprintf (_writer, "%c1", flag ? '+' : '-');
      for (ny::fv_it it = _fv.begin (); it != _fv.end (); ++it)
        std::fprintf (_writer, " %d:1", *it);
      std::fprintf (_writer, "\n");
    }
    // feature vector generators
    void _add_boolean_feature  (const bool flag)__attribute__((always_inline));
    void _add_boolean_feature  (const bool flag, const bool flag_)__attribute__((always_inline));
    void _add_string_feature   (const int &id)__attribute__((always_inline));
    void _add_string_feature   (const int &id, bool flag)__attribute__((always_inline));
    void _add_global_feature   (const int i, const int j, const Bunsetsu &bi, const Bunsetsu &bj);
    void _add_cluster_feature  (const Morph &m);
    void _add_cluster_feature  (const Bunsetsu &b, const Morph * ms);
    void _add_morpheme_feature (const int idx, const int offset);
    void _add_morpheme_feature (const int idx, const int offset, const bool flag);
    void _add_lexical_feature  (const Bunsetsu &b);
    void _add_lexical_feature  (const Bunsetsu &b, bool flag);
    void _add_context_feature  (const int from, const int to);
    void _event_gen_from_tuple (const int i);
    void _event_gen_from_tuple (const int i, const int j);
    void _event_gen_from_tuple (const int i, const int j, const int k);
    void _setMorphDic ();
    void _registerMorph (char * cs, const ssize_t &len, sbag_t &sbag,
                         std::set <ny::uint> &context_feature_ids);
    void _learn ();
#if defined (USE_OPAL) || defined (USE_MAXENT)
    void _processSample (const bool flag);
#endif
    template <const process_t MODE> void _analyze ();
    template <const process_t MODE> void _chunk   ();
    template <const process_t MODE> void _parse   ();
    template <const process_t MODE> void _parseLinear     ();
    template <const process_t MODE> void _parseChunking   ();
    template <const process_t MODE> void _parseBackward   ();
    template <const process_t MODE> void _parseTournament ();
    template <const input_t INPUT>  void _collectStat ();
    void _switch_classifier  (const input_t in);
    void _setup_learner      ();
    void _cleanup_learner    ();
    void _setup_classifier   (const input_t in, int argc, char ** argv);
    void _cleanup_classifier (const input_t in);
#ifdef USE_MAXENT
    void _project (ME_Sample &ms) const;
#endif
  public:
    parser (const option &opt) :
      _opt (opt), _pecco_opt (), _pecco_chunk (), _pecco_depnd (),
#ifdef USE_OPAL
      _opal_opt (), _opal (0), _ex_pool (),
#endif
#ifdef USE_SVM
      _tiny_param (), _tinysvm (0),
#endif
#ifdef USE_MAXENT
      _maxent_opt (), _libme (0),
#endif
      _s (0), _dict (0), _fi (1), _fv (), _context_feature_bits (0), _chunk_stat (), _depnd_stat (), _writer (0)
#ifdef USE_TIMER
      , _timer_pool ("J.DepP profiler"), _io_t (_timer_pool.push ("io")), _dict_t (_timer_pool.push ("dict")), _preproc_t (_timer_pool.push ("preproc", "sent.")), _chunk_t (_timer_pool.push ("chunk", "sent.")), _depnd_t (_timer_pool.push ("depnd", "sent.")), _classify_t (_timer_pool.push ("classify"))
#endif
    {}
    ~parser () { delete _dict; }
    // interface
    void      run         ();
    Sentence* setSentence () { return _s =  new Sentence (); }
  };
}
#endif /* PDEP_H */
